/*
 *  DriverLoader(tm) PCI module for Linux
 *  Copyright (c) 2003-2008 Linuxant inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define OS_NOSTRINGREDEFS

#include "oscompat.h"
#include "osservices.h"
#include "ndiswrap.h"
#include "netndis.h"

#include <linux/pci.h>
#include <linux/kmod.h>
#include <linux/ctype.h>

#ifndef HAVE_NETDEV_PRIV
#define netdev_priv(ndev) ((ndev)->priv)
#endif

MODULE_AUTHOR("Copyright (C) 2003-2004 Linuxant inc.");
MODULE_DESCRIPTION(DLDRDESC);
MODULE_LICENSE("GPL");
MODULE_INFO(supported, "yes");

static struct pci_device_id dldrpci_tbl[] =
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	{PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_CLASS_NETWORK_ETHERNET << 8, 0xffff00, 0},
	{PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_CLASS_NETWORK_OTHER << 8, 0xffff00, 0},
#else
	{0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0}, /* dummy non-matching entry for buggy depmod */
#endif
	{0,}                    /* terminate list */
};


MODULE_DEVICE_TABLE(pci, dldrpci_tbl);

extern struct semaphore dldr_init_mutex;

static int dldrpci_suspend(struct pci_dev *pdev, pm_message_t state)
{
	POS_DEVNODE pDevNode = pci_get_drvdata(pdev);
	struct net_device *ndev = pDevNode->netDev;
	struct netndis_private *priv = netdev_priv(ndev);
	ULONG wrapState;
	int err;
	
	printk(KERN_DEBUG "%s: entering sleep mode (state=%d)\n", ndev->name, GET_PM_MESSAGE_EVENT(state));

	wrapState = NdisWrapEnter();

	err = netndis_lock(priv, FALSE);
	if (err) {
		printk(KERN_ERR "%s: hw_unavailable on %s\n", ndev->name, __FUNCTION__);
		NdisWrapLeave(wrapState);
		return err;
	}

	err = __netndis_down(ndev);
	if (err)
		printk(KERN_WARNING "%s: %s: Error %d downing interface\n", ndev->name, __FUNCTION__, err);
	
	netif_device_detach(ndev);

	priv->hw_unavailable++;

	err = NdisWrapStatusToErrno(NdisWrapSetPowerState(priv->miniportAdapter, NDISWRAP_PWRSTATE_D0 + GET_PM_MESSAGE_EVENT(state)));
	if (err) {
		printk(KERN_WARNING "%s: transition to state D%d failed (err=%d)\n", ndev->name, GET_PM_MESSAGE_EVENT(state), err);
	} else {
		printk(KERN_DEBUG "%s: changed to state D%d\n", ndev->name, GET_PM_MESSAGE_EVENT(state));
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
	pci_save_state(pdev);
#else
	pci_save_state(pdev, priv->pci_state);
#endif
	pci_set_power_state(pdev, pci_choose_state(pdev, state));

	netndis_unlock(priv);

	NdisWrapLeave(wrapState);

	return 0;
}

static int dldrpci_resume(struct pci_dev *pdev)
{
	POS_DEVNODE pDevNode = pci_get_drvdata(pdev);
	struct net_device *ndev = pDevNode->netDev;
	struct netndis_private *priv = netdev_priv(ndev);
	ULONG wrapState;
	int err;

	printk(KERN_DEBUG "%s: waking up\n", ndev->name);

	wrapState = NdisWrapEnter();

	pci_set_power_state(pdev, 0);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
	pci_restore_state(pdev);
#else
	pci_restore_state(pdev, priv->pci_state);
#endif

	err = NdisWrapStatusToErrno(NdisWrapSetPowerState(priv->miniportAdapter, NDISWRAP_PWRSTATE_D0));
	if (err) {
		printk(KERN_WARNING "%s: transition to state D0 failed (err=%d)\n", ndev->name, err);
	} else {
		printk(KERN_DEBUG "%s: changed to state D0\n", ndev->name);
	}

	err = netndis_lock(priv, TRUE);
	if (err) {
		printk(KERN_ERR "%s: hw_unavailable on %s\n", ndev->name, __FUNCTION__);
		NdisWrapLeave(wrapState);
		return err;
	}

	netif_device_attach(ndev);

	priv->hw_unavailable--;

	if (priv->open && (! priv->hw_unavailable)) {
		err = __netndis_up(ndev);
		if (err)
			printk(KERN_WARNING "%s: %s: Error %d restarting interface\n", ndev->name, __FUNCTION__, err);
	}
	
	netndis_unlock(priv);

	NdisWrapLeave(wrapState);

	return 0;
}

static void __devexit dldrpci_remove (struct pci_dev *pdev)
{
	POS_DEVNODE pDevNode = pci_get_drvdata(pdev);
	struct net_device *ndev;
	ULONG wrapState;

	OsTracePrintf("%s: pDevNode=%p\n", __FUNCTION__, pDevNode);

	if(!pDevNode)
		return;

	down(&dldr_init_mutex);

	wrapState = NdisWrapEnter();

	ndev = pDevNode->netDev;
	pDevNode->netDev = NULL;

	if(ndev) {
		unregister_netdev(ndev);
	}

	NdisWrapHaltAdapter(pDevNode->miniportAdapter);

	pci_set_drvdata(pdev, NULL);

	if(ndev) {
		free_netdev(ndev);
	}
	NdisWrapDestroyAdapter(pDevNode->miniportAdapter);
	kfree(pDevNode);

	pci_release_regions(pdev);
	pci_disable_device(pdev);

	NdisWrapLeave(wrapState);

	up(&dldr_init_mutex);
}


static int __devinit dldrpci_probe (struct pci_dev *pdev,
		const struct pci_device_id *ent)
{
	int err;
	POS_DEVNODE pDevNode;
	struct net_device *ndev = NULL;
	char buf[100];
	char *argv[5], *envp[3];
	const char *p;
	u_int8_t csz;
	ULONG wrapState;
	UINT32 mask;

	OsTracePrintf("%s: pdev=%p\n", __FUNCTION__, pdev);

	pDevNode = kmalloc(sizeof(*pDevNode), GFP_KERNEL);
	if(!pDevNode) {
		return -ENOMEM;
	}

	down(&dldr_init_mutex);

	wrapState = NdisWrapEnter();

	memset(pDevNode, 0, sizeof(*pDevNode));
	pDevNode->hwDev = pdev;
	pDevNode->hwModule = THIS_MODULE;
	pDevNode->hwBusType = OSDEVTYPE_PCI;
	pDevNode->osPageOffset = PAGE_OFFSET;

	/* count : chars in slot_name to determine if fake domain_nr must be added
	 * to make 2.4 instance names identical to 2.6 */
	for(p = PCI_SLOT_NAME(pdev), csz = 0; *p; p++)
		if(*p == ':')
			csz++;
	snprintf(pDevNode->hwInstName, sizeof(pDevNode->hwInstName), "PCI-%s%s-%04X:%04X-%04X:%04X", csz == 1 ? "0000:" : "", PCI_SLOT_NAME(pdev), pdev->vendor, pdev->device, pdev->subsystem_vendor, pdev->subsystem_device);

	err = NdisWrapCreateAdapter(pDevNode, DLDRVARDIR, &pDevNode->miniportAdapter);
	OsTracePrintf("%s: NdisWrapCreateAdapter returned %d miniportAdapter=%p instname=%s pDevNode=%p\n", __FUNCTION__, err, pDevNode->miniportAdapter, pDevNode->hwInstName, pDevNode);
	if(err) {
		pDevNode->miniportAdapter = NULL;
		goto outerr;
	}

	if ( (err = pci_enable_device(pdev)) ) {
		printk(KERN_ERR "%s: pci_enable_device failed (err=%d)\n", PCI_SLOT_NAME(pdev), err);
		goto outerr;
	}

	if (pci_request_regions(pdev, (char*)(THIS_MODULE->name))) {
		printk(KERN_ERR "%s: pci_request_regions failed\n", PCI_SLOT_NAME(pdev)); 
		pci_disable_device(pdev);
		err = -ENODEV;
		goto outerr;
	}

	pci_set_power_state(pdev, 0);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
	pci_restore_state(pdev, NULL);
#endif

#if 0
	/* XXX enable memory-write-invalidate */
	if (pci_set_mwi(pdev)) {
		printk(KERN_ERR "%s: unable to set MWI\n", PCI_SLOT_NAME(pdev));
		pci_release_regions(pdev);
		pci_disable_device(pdev);
		err = -EIO;
		goto outerr;
	}
#endif

	NdisWrapGetDMAMask(pDevNode->miniportAdapter, &mask);

	/* XXX 32-bit addressing only */
	if (pci_set_dma_mask(pdev, mask)) {
		printk(KERN_ERR "%s: 32-bit DMA not available\n", PCI_SLOT_NAME(pdev));
		pci_release_regions(pdev);
		pci_disable_device(pdev);
		err = -EIO;
		goto outerr;
	}

	/*
	 * Cache line size is used to size and align various
	 * structures used to communicate with the hardware.
	 */
	pci_read_config_byte(pdev, PCI_CACHE_LINE_SIZE, &csz);
	if (csz == 0) {
		/*
		 * Linux 2.4.18 (at least) writes the cache line size
		 * register as a 16-bit wide register which is wrong.
		 * We must have this setup properly for rx buffer
		 * DMA to work so force a reasonable value here if it
		 * comes up zero.
		 */
		csz = L1_CACHE_BYTES / sizeof(u_int32_t);
		printk("%s: cache line size not set; forcing %u\n", PCI_SLOT_NAME(pdev), csz);
		pci_write_config_byte(pdev, PCI_CACHE_LINE_SIZE, csz);
	}
	/*
	 * The default setting of latency timer yields poor results,
	 * set it to the value used by other systems.  It may be worth
	 * tweaking this setting more.
	 */
	pci_write_config_byte(pdev, PCI_LATENCY_TIMER, 0xa8);

	pci_set_drvdata(pdev, pDevNode);

	ndev = alloc_netndisdev(pDevNode);
	if(!ndev) {
		pci_set_drvdata(pdev, NULL);
		pci_release_regions(pdev);
		pci_disable_device(pdev);
		err = -ENOMEM;
		goto outerr;
	}

#ifdef SET_MODULE_OWNER
	SET_MODULE_OWNER(ndev);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	SET_NETDEV_DEV(ndev, &pdev->dev);
#endif

	pDevNode->netDev = ndev;

	err = register_netdev(ndev);
	if (err) {
		printk(KERN_ERR "%s: Failed to register net device (%d)\n", ndev->name, err);
		pDevNode->netDev = NULL;
		free_netdev(ndev);
		pci_set_drvdata(pdev, NULL);
		pci_release_regions(pdev);
		pci_disable_device(pdev);
		goto outerr;
	}

	buf[0] = '\0';
	NdisWrapGetVendorDesc(pDevNode->miniportAdapter, buf, sizeof(buf));

	printk(KERN_DEBUG"%s:%s%s at %s (MAC address %02X:%02X:%02X:%02X:%02X:%02X) ready\n", ndev->name,
			buf[0] ? " ":"", buf,
			PCI_SLOT_NAME(pdev), 
			ndev->dev_addr[0], ndev->dev_addr[1], ndev->dev_addr[2],
			ndev->dev_addr[3], ndev->dev_addr[4], ndev->dev_addr[5]);

	NdisWrapLeave(wrapState);
	up(&dldr_init_mutex);

    envp [0] = "HOME=/";
    envp [1] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin:"DLDRSBINDIR;
    envp [2] = 0;

    argv[0] = DLDRSBINDIR"/dldrconfig";
    argv[1] = "--auto";
    sprintf(buf, "--netif=%s", ndev->name);
    argv[2] = buf;
    argv[3] = "--modalias";
    argv[4] = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    err = call_usermodehelper(argv[0], argv, envp);
#else
    err = call_usermodehelper(argv[0], argv, envp, 1);
#endif
	if(err && (err != SIGTERM)) {
        printk(KERN_ERR"%s: call_usermodehelper %s returned %d\n", __FUNCTION__, argv[0], err);
    }

	return 0;

outerr:
	if(pDevNode->miniportAdapter)
		NdisWrapDestroyAdapter(pDevNode->miniportAdapter);

	NdisWrapLeave(wrapState);
	up(&dldr_init_mutex);
	kfree(pDevNode);
	return err;
}

struct pci_driver dldrpci_driver = {
	.name           = "driverloader",
	.id_table       = dldrpci_tbl,
	.probe          = dldrpci_probe,
	.remove         = __devexit_p(dldrpci_remove),

	.suspend        = dldrpci_suspend,
	.resume         = dldrpci_resume,
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
/* Converts a hex string, optionally prefixed by "0x". If it's not a valid hex value, returns -1 */
static int32_t hex_string(const char *buf)
{
	int32_t value;
	int32_t digit;
	char c;

	value = 0;

	if (buf[0] == '0' && buf[1] == 'x')
		buf += 2;

	while (*buf) {
		c = *buf;

		if (c >= 'a' && c <= 'f')
			digit = c - 'a' + 10;
		else {
			if (c >= 'A' && c <= 'F')
				digit = c - 'A' + 10;
			else {
				if (c >= '0' && c <= '9')
					digit = c - '0';
				else
					return -1;
			}
		}

		value <<= 4;
		value |= digit;

		buf++;
	}

	return value;
}
#endif

static int __init dldrpci_init(void)
{
	int ret, err;
	char *argv[4], *envp[3];

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
	struct pci_dev *pdev;
	struct device *dev;
	FILE *infcache;
	uint16_t pdev_id[4];
	char c;
	int i;
	char buf[7]; /* strlen("0x1234") + 1 */
	int state;
	int is_pci_line;
	int32_t value;

	state = 0;
	i = 0;
	is_pci_line = 0;

	infcache = OsFOpen(DLDRCNFDIR "/.infcache", "r", NULL);
	if (!infcache) {
		OsErrorPrintf("%s: WARNING: Failed to open " DLDRCNFDIR "/.infcache\n", __FUNCTION__);
		goto register_pdev;
	}

	/* The format of this file is "file bus vendor device sub_vendor sub_device" and the 2 last values are optional */
	while (1) {
		err = OsFRead(&c, sizeof(c), 1, infcache, NULL);

		if (err != sizeof(c) && err != 0) {
			OsErrorPrintf("%s: Error reading " DLDRCNFDIR "/.infcache: %d\n", __FUNCTION__, err);
			break;
		}

		/* Whitespace or EOF */
		if (c == ' ' || c == '\t' || c == '\n' || err == 0) {
			/* End of word */
			if (i != 0) {
				buf[i] = 0;

				switch (state) {
				case 0:	/* File name we don't care about */
					state = 1;
					break;
				case 1: /* Bus type */
					if (strcmp(buf, "PCI"))
						state = -1;
					else {
						state = 2;
						is_pci_line = 1;
					}
					break;
				case 2 ... 5: /* One ID */
					/* Check that we have a valid 16-bit hex */
					value = hex_string(buf);
					if (value == -1) {
						is_pci_line = 0;
						state = -1;
					} else {
						pdev_id[state - 2] = value;
						state++;
					}
					break;
				default: /* Extra info on line or invalid line we don't care about */
					break;
				}

				i = 0;
			}

			/* End of line or EOF */
			/* state == 4 means that we got vendor:device and 6 means vendor:device and subvendor:subdevice */
			if ((c == '\n' || err == 0) && is_pci_line && (state == 4 || state == 6)) {
				pdev = NULL;

				while (1) {
					if (state == 4)
						pdev = pci_get_device(pdev_id[0], pdev_id[1], pdev);
					else
						pdev = pci_get_subsys(pdev_id[0], pdev_id[1], pdev_id[2], pdev_id[3], pdev);

					if (!pdev)
						break;

					if (pci_dev_driver(pdev) && (dev = get_device(&pdev->dev))) {
						printk(KERN_INFO "%s: device %04x:%04x is still driven by %s, trying to release...\n", __FUNCTION__, pdev_id[0], pdev_id[1], dev->driver->owner->name);
						device_release_driver(dev);
						put_device(dev);
					}
				}

				is_pci_line = 0;
				state = 0;
			}
		} else
			/* Other, if no overflow, fill buffer */
			if (i != 6)
				buf[i++] = c;

		/* EOF */
		if (err == 0)
			break;
	}

	OsFClose(infcache);

register_pdev:
#endif
	ret = PCI_REGISTER_DRIVER(&dldrpci_driver);

	if (ret < 0)
		return ret;

	envp[0] = "HOME=/";
	envp[1] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin:"DLDRSBINDIR;
	envp[2] = 0;
	argv[0] = DLDRSBINDIR"/dldrconfig";
	argv[1] = "--auto";
	argv[2] = "--mod-init";
	argv[3] = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	err = call_usermodehelper(argv[0], argv, envp);
#else
	err = call_usermodehelper(argv[0], argv, envp, 1);
#endif
	if (err && (err != SIGTERM)) {
		printk(KERN_ERR"%s: call_usermodehelper %s returned %d\n", __FUNCTION__, argv[0], err);
	}

	return ret;
}

static void __exit dldrpci_cleanup(void)
{
	NdisWrapNotifyBusDeregister(OSDEVTYPE_PCI);
	pci_unregister_driver(&dldrpci_driver);
}

module_init(dldrpci_init)
module_exit(dldrpci_cleanup)

